@@ -44,8 +44,8 @@ class Markov(object):
             cache[key] = [w]
 
     def feed(self, message):
-        splitted = message.split(' ')
-        for w1, w2, w3 in self.triples(self.words[-1:] + splitted + ['\n']):
+        splitted = list(map(sys.intern, message.split(' ')))
+        for w1, w2, w3 in self.triples(self.words[-2:] + splitted + ['\n']):
             self._add_key_to_cache((w1, w2), self.forward_cache, w3)
             self._add_key_to_cache((w3, w2), self.backward_cache, w1)
         self.words.extend(splitted + ['\n'])
@@ -66,11 +66,8 @@ class Markov(object):
             # select one of them, and take the next word.
             possible_indexes = [i+1 for (i, x) in enumerate(self.words[1:-1])
                               if self.words[i+1] == seed_word]
-            next_word = self.words[random.choice(possible_indexes)+d]
-            if backward:
-                assert (seed_word, next_word) in self.backward_cache
-            else:
-                assert (seed_word, next_word) in self.forward_cache
+            index = random.choice(possible_indexes)
+            next_word = self.words[index+d]
         else:
             raise ValueError('%s is not in the corpus.' % seed_word)
         return (seed_word, next_word)
@@ -83,10 +80,13 @@ class Markov(object):
             w1, w2 = '\n', seed_word
         else:
             w1, w2 = seed_word, next_word
+        (w1, w2) = sys.intern(w1), sys.intern(w2)
         gen_words = []
         for i in range(max_size):
             gen_words.append(w1)
             new = '\n'
+            if (w1, w2) not in cache:
+                break
             new = random.choice(cache[(w1, w2)])
             if new == '\n':
                 break
@@ -99,15 +99,16 @@ class Markov(object):
 
 def mixed_encoding_extracting(f):
     def newf(msg):
-        encoding = 'utf8'
-        if chardet:
-            encoding = chardet.detect(msg)['encoding']
         try:
-            msg = msg.decode(encoding)
+            msg = msg.decode()
         except UnicodeDecodeError:
-            return None
-        else:
-            return f(msg)
+            if chardet:
+                encoding = chardet.detect(msg)['encoding']
+                try:
+                    msg = msg.decode(encoding)
+                except UnicodeDecodeError:
+                    return None
+        return f(msg)
     return newf
 
 REGEXPS = {
